/*
    i2c-MachZ.c - Linux kernel module for hardware monitoring
    Copyright (c) 2001 Hjalmar Hevers ,hhevers@ste-el.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
    This driver supports the ACCESS.BUS-Device on the MachZ from ZFLinux-Devices
*/


#include <linux/version.h>
#include <linux/module.h>
#include <asm/io.h>
#include <asm/ioctl.h>
#include <linux/kernel.h>
#include <linux/stddef.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/vmalloc.h>

/* MachZ-dependent Registers */
#define S_IO_BASEINDEXADR  0x2E
#define S_IO_BASEDATAADR   0x2F
#define S_IO_LOGICDEVICEOFFSET 0x07
#define S_IO_DEVICECONTROL 0x30
#define S_IO_DEVICEPORTADRESS  0x60
#define S_IO_DEVICEINTERRUPT   0x70
#define S_IO_DEVICEINTTYPE   0x71

#define S_IO_DEVENABLE  0x01
#define S_IO_DEVDISABLE  0
#define S_IO_SMBINTTYPE  0x02
#define S_IO_SMBINTERRUPT 0x07

#define SMB_LDN 0x08
#define SMB_BASE 0x300
#define SMB_DATAOFFSET 0
#define SMB_STATUSOFFSET 1
#define SMB_CONSTATOFFSET 2
#define SMB_CONTROL1OFFSET 3
#define SMB_OWNADROFFSET 4
#define SMB_CONTROL2OFFSET 5


/* Other settings */
#define MAX_TIMEOUT 50
#define SMB_MAJOR 89
#define SMBSTATINIT 1
#define SMBSTATOPEN 2
#define SMBSTATBUSY 3
#define SMBSTATREADY 4

/* SMB_MachZ ioctl command constants */

long long i2c_lseek(struct file * file, long long offset, int origin) {	return -ESPIPE; };
int i2c_open(struct inode * inode, struct file * file);
int i2c_release (struct inode * inode, struct file * file);
ssize_t i2c_write(struct file * file,const char * buf, size_t count, loff_t *ppos);
ssize_t i2c_read(struct file * file,char * buf, size_t count, loff_t *ppos) ;
int i2c_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg);

struct file_operations i2c_fops = {
	i2c_lseek,
	i2c_read,
	i2c_write,
	NULL,  			/* i2c_readdir	*/
	NULL,			/* i2c_select 	*/
	i2c_ioctl,
	NULL,			/* i2c_mmap 	*/
	i2c_open,
	NULL,			/* i2c_flush	*/
	i2c_release,
};

static struct smb_dev {
  unsigned char r_w_adress;
  unsigned int status;
  unsigned int timeout;
  unsigned long count;
  unsigned long maxcount;
  unsigned char *datapointer;
} smb_device = {0,0,}; 

struct smb_channel {
  unsigned char adress;
  int timeout ;
};
  
/* insmod parameters */

static int force_addr = 0;
MODULE_PARM(force_addr, "i");
MODULE_PARM_DESC(force_addr,
		 "Forcibly enable the SMB_MachZ at the given address. ");
static int force_int = 0;
MODULE_PARM(force_int , "i");
MODULE_PARM_DESC(force_int ,
		 "Forcibly put the SMB_MachZ at the given Interrupt. ");
#ifdef MODULE
static
#else
extern
#endif
int __init i2c_SMB_MachZ_init(void);
static int __init SMB_MachZ_cleanup(void);
static int SMB_MachZ_setup(void);
void SmbIRQhd (int, void * ,struct pt_regs *);

#ifdef MODULE
extern int init_module(void);
extern int cleanup_module(void);
#endif				/* MODULE */


static unsigned short SMB_MachZ_smba = 0;
static unsigned short SMB_MachZ_int = 0;


int SMB_MachZ_setup(void)
{
	int error_return = 0;
       	if (force_addr) {
		SMB_MachZ_smba = force_addr & 0xfff0;
	} else {
		SMB_MachZ_smba = SMB_BASE;
	}

	if (force_int) {
		SMB_MachZ_int = force_int & 0x0F;
	} else {
	  SMB_MachZ_int =  S_IO_SMBINTERRUPT & 0x0F;
	}
	if (check_region(SMB_MachZ_smba, 6)) {
	  printk
	    ("i2c-SMB_MachZ.o: SMB_MachZ_smb region 0x%x already in use!\n",
	     SMB_MachZ_smba);
	  error_return = -ENODEV;
	  goto END;
	}
	if( request_irq(SMB_MachZ_int,SmbIRQhd,0,"SMB",NULL) !=0){
	  printk
	    ("i2c-SMB_MachZ.o: SMB_MachZ_int 0x%x already in use!\n",
	     SMB_MachZ_int);
	  error_return = -ENODEV;
	  goto END;
	}
	/* We program the new address here. Just to make
	   sure, we disable the SMB_MachZ first. */
	outb_p(   S_IO_LOGICDEVICEOFFSET, S_IO_BASEINDEXADR   );
	outb_p(   SMB_LDN ,  S_IO_BASEDATAADR  );
	outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
	outb_p(   S_IO_DEVDISABLE,  S_IO_BASEDATAADR    );
	outb_p(   S_IO_DEVICEPORTADRESS+1, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)(SMB_MachZ_smba & 0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEPORTADRESS, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)((SMB_MachZ_smba>>8) & 0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEINTERRUPT, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)(SMB_MachZ_int&0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEINTTYPE, S_IO_BASEINDEXADR );
	outb_p(   S_IO_SMBINTTYPE, S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
	outb_p(   S_IO_DEVENABLE, S_IO_BASEDATAADR);
	outb_p(0, SMB_MachZ_smba+SMB_OWNADROFFSET   );
	outb_p(((54<<1)+1), SMB_MachZ_smba+SMB_CONTROL2OFFSET   );
	request_region(SMB_MachZ_smba, 6,"SMB");
 END:
	return error_return;
}

int __init i2c_SMB_MachZ_init(void)
{
	int ret;
	if ((ret = SMB_MachZ_setup())) {
		printk
		    ("i2c-SMB_MachZ.o: SMB_MachZ-Module not inserted.\n");
		SMB_MachZ_cleanup();
		return ret;
	}
	printk("i2c-SMB_MachZ.o: SMB_MachZ bus initialized\n");
	if (register_chrdev(SMB_MAJOR,"smb",&i2c_fops)) {
		printk("i2c: unable to get major %d for i2c bus\n",SMB_MAJOR);
	}
	smb_device.status=SMBSTATINIT;
	return 0;
}

int __init SMB_MachZ_cleanup(void)
{
  outb_p( (54<<1) ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb disable
  outb_p(   S_IO_LOGICDEVICEOFFSET, S_IO_BASEINDEXADR );
  outb_p(     SMB_LDN,  S_IO_BASEDATAADR );
  outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
  outb_p(     S_IO_DEVDISABLE, S_IO_BASEDATAADR );
  if (smb_device.status >=  SMBSTATINIT) {
    release_region(SMB_MachZ_smba, 6);
    free_irq(SMB_MachZ_int,NULL);
    smb_device.status =  0;
  }
  unregister_chrdev(SMB_MAJOR,"smb");
  return 0;
}

EXPORT_NO_SYMBOLS;

#ifdef MODULE

MODULE_AUTHOR
    ("Hjalmar Hevers > hhevers@ste-el.de");
MODULE_DESCRIPTION("SMB_MachZ SMBus driver");

int init_module(void)
{
	return i2c_SMB_MachZ_init();
}

int cleanup_module(void)
{
	return SMB_MachZ_cleanup();
}

#endif				/* MODULE */

void SmbIRQhd (int irq,void *dev_id ,struct pt_regs *reg_pnt){

  if(inb_p(SMB_MachZ_smba+SMB_STATUSOFFSET)& 0x30){//Buserror or Negack
    outb_p(0x0,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable
    outb_p(0x02,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable; Stop-Condition
    outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);
    smb_device.status=SMBSTATREADY;
    return;
  }
  smb_device.timeout=0;
  switch (smb_device.status){
  case SMBSTATOPEN:
    smb_device.status=SMBSTATBUSY;
    outb_p(smb_device.r_w_adress,SMB_MachZ_smba);//Adress_r_w - out
    break;
  case SMBSTATBUSY:
    if(smb_device.r_w_adress & 1){//read
      if(smb_device.count >= smb_device.maxcount-1){
	outb_p(0,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable
	outb_p(0x02,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable; Stop-Condition
	outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);	
	smb_device.status=SMBSTATREADY;
      }
      else if(smb_device.count >= smb_device.maxcount-2){//last data
	outb_p(0x14,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//NACK
      }
      *(smb_device.datapointer+smb_device.count++)=inb_p(SMB_MachZ_smba);
    }
    else{//write
      if(smb_device.count>=smb_device.maxcount){//last data
	outb_p(0x0,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable
	outb_p(0x02,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable; Stop-Condition
	outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);	
	smb_device.status=SMBSTATREADY;
	return;
      }
      outb_p(*(smb_device.datapointer+smb_device.count++),SMB_MachZ_smba);//Data - out
    };
    break;
  default:
    break;
  }
  return;
};

int i2c_open(struct inode * inode, struct file * file){
  struct smb_channel *channelp;
  channelp = vmalloc(sizeof(struct smb_channel));
  file->private_data = channelp;
  switch(MINOR(inode->i_rdev)){
  case 0:
    channelp->adress=0x50;
    break;
  case 1:
     channelp->adress=0x30;
    break;
  case 2:
    channelp->adress=0x31;
    break;
  case 3:
    channelp->adress=0x32;
    break;
  case 4:
    channelp->adress=0x33;
    break;
  case 5:
    channelp->adress=0x34;
    break;
  default:
    channelp->adress=0;
    break;
  };
  MOD_INC_USE_COUNT;
  return (0);
};

int i2c_release (struct inode * inode, struct file * file){
  vfree( file->private_data );
  MOD_DEC_USE_COUNT;
  return(0);
};

ssize_t i2c_write(struct file * file,const char * buf, size_t count, loff_t *ppos){
  unsigned char temp;
  ((struct smb_channel *)file->private_data)->timeout=0;
  while(smb_device.status!=SMBSTATINIT && ((struct smb_channel *)file->private_data)->timeout++ < MAX_TIMEOUT){
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  }
  if (((struct smb_channel *)file->private_data)->timeout >= MAX_TIMEOUT)return(0);
  smb_device.status = SMBSTATOPEN;
  if(inb_p(SMB_MachZ_smba+ SMB_CONSTATOFFSET)&0x02)outb_p(0x02,SMB_MachZ_smba+ SMB_CONSTATOFFSET);
  ((struct smb_channel *)file->private_data)->timeout=0;
  temp= inb_p(SMB_MachZ_smba + SMB_CONSTATOFFSET);
  while(( ((inb_p(SMB_MachZ_smba+ SMB_CONSTATOFFSET)&0x10)==0) ||
	 ((temp & 0x01)==1)) && ((struct smb_channel *)file->private_data)->timeout++ < MAX_TIMEOUT){
    if((temp & 0x01)==1){
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(1);
      temp= inb_p(SMB_MachZ_smba + SMB_CONSTATOFFSET);
      if((temp & 1)==0) continue ;
    }
    outb_p( (54<<1) ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb disable
    outb_p( (54<<1)+1 ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb enable
    outb_p(0x22,SMB_MachZ_smba+ SMB_CONSTATOFFSET);
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  }
  if (((struct smb_channel *)file->private_data)->timeout >= MAX_TIMEOUT){
    smb_device.status = SMBSTATINIT;
    return(0);
  }
  outb_p(0x30,SMB_MachZ_smba+SMB_STATUSOFFSET);//all clear
  smb_device.r_w_adress= ((struct smb_channel *)file->private_data)->adress<<1;
  smb_device.datapointer=(unsigned char *)vmalloc(count);
  copy_from_user(smb_device.datapointer,buf,count);
  smb_device.timeout=0;
  smb_device.count=0;
  smb_device.maxcount=count;
  outb_p(0x04,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr enable
  outb_p(0x05,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr enable make startcondition
  while( smb_device.status !=SMBSTATREADY && smb_device.timeout++ <  MAX_TIMEOUT){
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  }
  if(smb_device.timeout >=  MAX_TIMEOUT){
    outb_p(0x02,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable; Stop-Condition
    outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);//all clear
  }
  vfree( smb_device.datapointer);
  smb_device.status =SMBSTATINIT;
  return(smb_device.count);
};

ssize_t i2c_read(struct file * file,char * buf, size_t count, loff_t *ppos){
  unsigned char temp;
  ((struct smb_channel *)file->private_data)->timeout=0;
  while(smb_device.status!=SMBSTATINIT && ((struct smb_channel *)file->private_data)->timeout++ < MAX_TIMEOUT){
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  };
  if (((struct smb_channel *)file->private_data)->timeout >= MAX_TIMEOUT)return(0);
  smb_device.status = SMBSTATOPEN;
  if(inb_p(SMB_MachZ_smba+ SMB_CONSTATOFFSET)&0x02)outb_p(0x02,SMB_MachZ_smba+ SMB_CONSTATOFFSET);
  ((struct smb_channel *)file->private_data)->timeout=0;

  temp= inb_p(SMB_MachZ_smba + SMB_CONSTATOFFSET);
  while( ( ((inb_p(SMB_MachZ_smba+ SMB_CONSTATOFFSET)&0x10)==0) ||
	  ((temp & 0x01)==0x01)) && ((struct smb_channel *)file->private_data)->timeout++ < MAX_TIMEOUT){
    if((temp & 0x01) == 0x01){
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(1);
      temp= inb_p(SMB_MachZ_smba + SMB_CONSTATOFFSET);
      if((temp & 0x01)==0) continue ;
    }
    outb_p( (54<<1) ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb disable
    outb_p( (54<<1)+1 ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb enable
    outb_p(0x22,SMB_MachZ_smba+ SMB_CONSTATOFFSET);
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  }
  if (((struct smb_channel *)file->private_data)->timeout >= MAX_TIMEOUT){
    smb_device.status = SMBSTATINIT;
    return(0);
  };
  outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);//all Errors clear
  smb_device.r_w_adress=( ((struct smb_channel *)file->private_data)->adress<<1) + 1;
  smb_device.datapointer=(unsigned char *)vmalloc(count);
  smb_device.timeout=0;
  smb_device.count=0;
  smb_device.maxcount=count;
  outb_p(0x04,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr enable
  outb_p(0x05,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr enable make startcondition
  while( smb_device.status !=SMBSTATREADY && smb_device.timeout++ <  MAX_TIMEOUT){
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
  };
  if(smb_device.timeout >=  MAX_TIMEOUT){
    outb_p(0x02,SMB_MachZ_smba+SMB_CONTROL1OFFSET);//interr disable; Stop-Condition
     outb_p(0x70,SMB_MachZ_smba+SMB_STATUSOFFSET);//all Errors clear
  };
  copy_to_user(buf,smb_device.datapointer,smb_device.count);
  vfree( smb_device.datapointer);
  smb_device.status =SMBSTATINIT;
  return(smb_device.count);
} ;

int i2c_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg){
  switch(cmd){
  case 0x654 /* SMB_MachZ_SETADRESS */ :
     ((struct smb_channel *)file->private_data)->adress=(unsigned char)(arg & 0x7F);
    return(0);
  case 0x655 /*  SMB_MachZ_GETREG */:
    return(inb_p(SMB_MachZ_smba+((arg>>8) &0x0F)));
  case 0x656 /*  SMB_MachZ_SETREG */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+((arg>>8) &0x0F));
    return(0);
  default:
    return(0);
  };
};




